///////////////////////////////////////////////////////////////////////////////
//
// skeleton.h
//
// Purpose:   Declaration of classes related to an articulated figure
//            (or skeleton) model.
//            Classes: Joint.
//
// Created:   Jaroslav Semancik, 13/10/2003
//
///////////////////////////////////////////////////////////////////////////////

#ifndef SKELETON_H
#define SKELETON_H

///////////////////////////////////////////////////////////////////////////////
//
// class Joint
//
// Class Joint represents a joint in an articulated figure (or skeleton).
// Joints are arranged in a tree-like hierarchy. Position and orientation of
// each joint is relative to its parent in the hierarchy. Any of position or
// rotation coordinates may be a degree-of-freedom - it varies in time. Then
// its actual value for a particular time depends on a particular motion
// applied to the skeleton.
//
///////////////////////////////////////////////////////////////////////////////


class Joint
{
public:

    string name;        // name of the joint

// ? -> to private ?
    Joint *parent;      // pointer to a parent joint
    int n_children;     // number of children in a skeleton hierarchy
    Joint **child;      // array of pointers to childen in the hierarchy


    // Creates a joint with name nam, rest position rest_pos, rest rotation
    // rest_rot and degree-of-freedom indices to motion table for position and
    // rotation in pos_dof, rot_dof respectively. The joint is created with no
    // children and no parent. They must be added by AddChild method.

    Joint(const string& nam, const Vector& rest_pos, const Vector& rest_rot,
          const IntVector& pos_dof, const IntVector& rot_dof) throw();


    // Destructor - destroys recursively all joint's descendants first.

    ~Joint() throw();


    // Add a joint chld as a direct descendant of the joint in the hierarchy
    // and set the joint as a perent of the child.

    void AddChild(Joint* chld) throw();


    // Draw the joint and its subhierarchy in a given motion and time.

    void Draw(const Motion *motion, double time);


    // Export the joint and its subhierarchy in a given motion and time
    // to a file as POV-Ray blob components.

    void Export(ofstream& file, const Motion *motion, double time);


private:

    // Position and rotation of the joint in a rest pose.

    Vector rest_position;   // position in the parent's coordinate frame
    Vector rest_rotation;   // orientation relative to parent's rotation
                            // as Euler angles - rotations around x, y and z axis

    // The following coordinates are indices of the respective DOF signals in
    // the motion table for all degrees of freedom in the joint. Those signals
    // overide the rest pose values. Coordinate value -1 means that it is not
    // a degree of freedom.

    IntVector position_dof; // indices of DOFs for position
    IntVector rotation_dof; // indices of DOFs for rotation

    void eval_instant_position(Vector &result, const Motion *motion,
                               double time);
    void eval_instant_rotation(Vector &result, const Motion *motion,
                               double time);
};



///////////////////////////////////////////////////////////////////////////////
//
// class Shot
//
// Class Shot represents an animation of one character performing a motion.
// It associates a particular skeleton with a particular motion.
// Responsibility to associate a compatible skeleton-motion pair lies on the
// user of the class.
//
///////////////////////////////////////////////////////////////////////////////

class Shot
{
public:

    string name;                // name of the shot

    Joint  *skeleton;           // root of the skeleton
    Motion *motion;             // motion applied to the skeleton

    unsigned char r, g, b;      // color of the skeleton


    // Creates a shot of name nam with skeleton skel of rgb color (rr, gg, bb)
    // associated with motion mot.

    Shot(const string& nam, Joint* skel, Motion* mot,
        unsigned char rr, unsigned char gg, unsigned char bb) throw();


    // Destructor - destroys associated skeleton and motion.

    ~Shot();


    // Draw the shot in a given time and DOF wrapping state.

    void Draw(double time, bool wrapping);
    
    
    // Export the shot in a given time to a given file.

    void Export(ofstream& file, double time);
};



///////////////////////////////////////////////////////////////////////////////
//
// global variables related to articulated figures
//
///////////////////////////////////////////////////////////////////////////////


// array of all skeletons

// extern "C++" vector<Joint*> skeletons;


// array of all shots

extern "C++" vector<Shot*> shots;

#endif

